from fastapi import FastAPI, HTTPException, status
from pydantic import BaseModel
import time
from . import db, cache, queue

class KVPut(BaseModel):
    value: str

app = FastAPI(title="Distributed KV Store")

@app.on_event("startup")
async def startup():
    await cache.init()
    db.init_sync()
    queue.init()

@app.get("/")
def health():
    return {"status": "ok"}

@app.get("/kv/{key}")
async def get_key(key: str):
    if (v := await cache.get(key)) is not None:
        return {"data": {"value": v, "source": "redis"}}

    rec = db.get(key)
    if rec:
        await cache.set(key, rec.value)
        return {"data": {"value": rec.value, "source": "db"}}
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Key not found")

@app.put("/kv/{key}", status_code=status.HTTP_202_ACCEPTED)
async def put_key(key: str, body: KVPut):
    ts = int(time.time_ns())
    queue.publish("add_key", {"key": key, "value": body.value, "ts": ts})
    return {"queued": True, "ts": ts}

@app.delete("/kv/{key}", status_code=status.HTTP_202_ACCEPTED)
async def delete_key(key: str):
    ts = int(time.time_ns())
    queue.publish("del_key", {"key": key, "ts": ts})
    return {"queued": True, "ts": ts}
