import json, time, asyncio, backoff, os
import aioredis, pika
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from api.app.models import Base, KV

REDIS_HOST = os.getenv("REDIS_HOST", "redis")
RABBIT_URL = os.getenv("RABBIT_URL", "amqp://guest:guest@rabbit1:5672/")
DB_DSN = os.getenv("DB_DSN", "sqlite:///data/app.db")
QUEUE_ADD = "add_key"
QUEUE_DEL = "del_key"

redis = None
engine = create_engine(DB_DSN, connect_args={"check_same_thread": False})
Base.metadata.create_all(engine)

@backoff.on_exception(backoff.expo, Exception, max_tries=5)
async def init_redis():
    global redis
    redis = await aioredis.from_url(f"redis://{REDIS_HOST}", decode_responses=True)

asyncio.run(init_redis())

def upsert(session: Session, key: str, value: str, ts: int):
    rec = session.get(KV, key)
    if rec is None or rec.last_updated <= ts:
        if rec is None:
            rec = KV(key=key, value=value, last_updated=ts)
            session.add(rec)
        else:
            rec.value = value
            rec.last_updated = ts
        session.commit()
        return True
    return False

def delete(session: Session, key: str, ts: int):
    rec = session.get(KV, key)
    if rec and rec.last_updated <= ts:
        session.delete(rec)
        session.commit()
        return True
    return False

def handle_add(ch, method, properties, body):
    payload = json.loads(body)
    key, value, ts = payload["key"], payload["value"], payload["ts"]
    with Session(engine) as s:
        if upsert(s, key, value, ts):
            asyncio.run(redis.set(key, value))
    ch.basic_ack(delivery_tag=method.delivery_tag)

def handle_del(ch, method, properties, body):
    payload = json.loads(body)
    key, ts = payload["key"], payload["ts"]
    with Session(engine) as s:
        if delete(s, key, ts):
            asyncio.run(redis.delete(key))
    ch.basic_ack(delivery_tag=method.delivery_tag)

def main():
    connection = pika.BlockingConnection(pika.URLParameters(RABBIT_URL))
    channel = connection.channel()
    channel.queue_declare(QUEUE_ADD, durable=True)
    channel.queue_declare(QUEUE_DEL, durable=True)
    channel.basic_qos(prefetch_count=10)
    channel.basic_consume(QUEUE_ADD, handle_add)
    channel.basic_consume(QUEUE_DEL, handle_del)
    channel.start_consuming()

if __name__ == "__main__":
    main()
